﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Shuffle;

namespace BasicMaze
{
    /// <summary>
    /// Directional contstants.
    /// </summary>
    public enum Directions { N, E, S, W };

    /// <summary>
    /// Means for creating a basic two-dimensional maze.
    /// </summary>
    public class Maze
    {
        //array of Cells that make up the Maze
        private Cell[,] _mazeCells;
        //start and finish positions of Maze
        private int _xFinish;
        private int _yFinish;
        private int _xStart;
        private int _yStart;

        /// <summary>
        /// Create a new Maze.
        /// </summary>
        /// <param name="xCells">Number of cells in the vertical/X dimension.</param>
        /// <param name="yCells">Number of cells in the horizontal/Y dimension.</param>
        public Maze(int xCells, int yCells)
        {
            int cellCount = xCells * yCells;//how many cells are in this maze?
            int distanceCount = 0;//keep track of how far we are from maze creation origin
            int distanceHigh = distanceCount;//keep track of the farthest distance

            //set up the grid of cells
            _mazeCells = new Cell[xCells, yCells];
            //instantiate all cells
            for (int i = 0; i < xCells; i++)
            {
                for (int j = 0; j < yCells; j++)
                {
                    _mazeCells[i, j] = new Cell(i, j);
                }
            }
            
            //pick a random starting place for maze generation
            //it will be the exit
            Random rnd = new Random();
            _xFinish = rnd.Next(xCells);
            _yFinish = rnd.Next(yCells);
            int currentX = _xFinish;
            int currentY = _yFinish;

            //ways we can go (refreshed for each Cell)
            List<int> directionsOpen = new List<int>();
            //pile of positions for this stack-based maze generator
            List<Position> positionStack = new List<Position>();

            //load in the starting position
            positionStack.Add(new Position(currentX,currentY));
            _mazeCells[currentX, currentY].DistanceFromStart = distanceCount;
            cellCount--;

            //while there are cells yet to visit (forces a "full maze" with no inaccessable cells)
            while (cellCount > 0)
            {
                directionsOpen.Clear();

                //find directions to places not visited
                if (currentX > 0 && !_mazeCells[currentX - 1, currentY].anyOpenPaths())
                {
                    directionsOpen.Add((int)Directions.N);                                    
                }
                if (currentY < yCells - 1 && !_mazeCells[currentX, currentY + 1].anyOpenPaths())
                {
                    directionsOpen.Add((int)Directions.E);                                    
                }
                if (currentX < xCells - 1 && !_mazeCells[currentX + 1, currentY].anyOpenPaths())
                {
                    directionsOpen.Add((int)Directions.S);
                }
                if (currentY > 0 && !_mazeCells[currentX, currentY - 1].anyOpenPaths())
                {
                    directionsOpen.Add((int)Directions.W);                
                }

                //if there is a direction to take:
                if (directionsOpen.Count > 0)
                {
                    //shuffle them
                    directionsOpen.Shuffle();
                    //pick first one, "go there" while knocking down walls between them
                    switch (directionsOpen.ElementAt(0))
                    {
                        case (int)Directions.N:
                            _mazeCells[currentX, currentY].setOpenPath((int)Directions.N);
                            currentX--;
                            _mazeCells[currentX, currentY].setOpenPath((int)Directions.S);
                            break;
                        case (int)Directions.E:
                            _mazeCells[currentX, currentY].setOpenPath((int)Directions.E);
                            currentY++;
                            _mazeCells[currentX, currentY].setOpenPath((int)Directions.W);
                            break;
                        case (int)Directions.S:
                            _mazeCells[currentX, currentY].setOpenPath((int)Directions.S);
                            currentX++;
                            _mazeCells[currentX, currentY].setOpenPath((int)Directions.N);
                            break;
                        case (int)Directions.W:
                            _mazeCells[currentX, currentY].setOpenPath((int)Directions.W);
                            currentY--;
                            _mazeCells[currentX, currentY].setOpenPath((int)Directions.E);
                            break;
                    }
                    //increment distance
                    distanceCount++;
                    _mazeCells[currentX, currentY].DistanceFromStart = distanceCount;
                    if (distanceCount > distanceHigh)
                    {
                        distanceHigh = distanceCount;                    
                    }

                    //add position
                    positionStack.Add(new Position(currentX, currentY));

                    //knock down cell count by one
                    cellCount--;
                }

                //else, when there are no new paths to forge:
                else 
                {
                    //go back - pop cell off the list
                    positionStack.RemoveAt(positionStack.Count - 1);
                    currentX = positionStack.ElementAt(positionStack.Count - 1).XPosition;
                    currentY = positionStack.ElementAt(positionStack.Count - 1).YPosition;
                    distanceCount--;
                }
            }

            //which cell is the farthest from the start?
            //(or just get the last one we find, if a tie)
            _xStart = _xFinish;
            _yStart = _yFinish;
            for (int i = 0; i < xCells; i++)
            {
                for (int j = 0; j < yCells; j++)
                {
                    if (_mazeCells[i, j].DistanceFromStart == distanceHigh)
                    {
                        _xStart = i;
                        _yStart = j;
                    }
                }
            }
        }

        /// <summary>
        /// Find vertical/X position of finish point (where maze was generated from).
        /// </summary>
        /// <returns>Vertical/X position of finish point.</returns>
        public int getXFinish()
        {
            return _xFinish;        
        }

        /// <summary>
        /// Find horizontal/Y position of finish point (where maze was generated from).
        /// </summary>
        /// <returns>Horizontal/Y position of finish point.</returns>
        public int getYFinish()
        {
            return _yFinish;
        }

        /// <summary>
        /// Find vertical/X position of starting point (among furthest from finish).
        /// </summary>
        /// <returns>Vertical/X position of starting point.</returns>
        public int getXStart()
        {
            return _xStart;
        }

        /// <summary>
        /// Find horizontal/Y position of starting point (among furthest from finish).
        /// </summary>
        /// <returns>Horizontal/Y position of starting point.</returns>
        public int getYStart()
        {
            return _yStart;
        }

        /// <summary>
        /// Get the array of Cells that make up the Maze.
        /// </summary>
        /// <returns>Two-dimensional array of maze Cells.</returns>
        public Cell[,] getCells()
        {
            return _mazeCells;        
        }
    }
}
