﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using BasicMaze;

namespace Maize
{
    /// <summary>
    /// Interface to show and play a Maize game.
    /// </summary>
    public partial class MaizeGameForm : Form
    {
        //cells that make up the maze
        private Cell[,] _mazeCells;
        //start, finish, dimensions
        private int _xStartPos;
        private int _yStartPos;
        private int _xFinishPos;
        private int _yFinishPos;
        private int _xSetSize = 10; //defaults
        private int _ySetSize = 10;
        //where the player is
        private int _playerX;
        private int _playerY;
        //winning state (congratulatory message drawn when true)
        private bool _winning = false;

        //means to draw the maze
        MazeDrawing _myMazeDrawing;

        /// <summary>
        /// Creates new instance of the Maize interface
        /// </summary>
        public MaizeGameForm()
        {
            InitializeComponent();
            this.WindowState = FormWindowState.Maximized;//start embiggened
            SummonMaze(_xSetSize, _ySetSize);//call up a (default-dimensioned) maze
        }

        //calls up a maze: vertical extent, horizontal extent
        private void SummonMaze(int xSize, int ySize)
        {
            //construct a maze
            Maze myMaze = new Maze(xSize, ySize);

            //get its start/finish particulars
            _xStartPos = myMaze.getXStart();
            _yStartPos = myMaze.getYStart();
            _xFinishPos = myMaze.getXFinish();
            _yFinishPos = myMaze.getYFinish();

            //get its cells
            _mazeCells = myMaze.getCells();

            //initialize game pieces
            _playerX = _xStartPos;
            _playerY = _yStartPos;
            _mazeCells[_xStartPos, _yStartPos].PlayerMarker = true;

            //fire up maze drawing class
            _myMazeDrawing = new MazeDrawing(_mazeCells, xSize, ySize, _xStartPos, _yStartPos, _xFinishPos, _yFinishPos);
        }

        //Show the AboutBox.
        private void aboutMaizeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            screenTimer.Stop();
            _winning = false;
            MaizeAboutBox aboutBox = new MaizeAboutBox();
            aboutBox.ShowDialog();
        }

        //close the application
        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        //show a dialog to specify size for a new maze; if result OK, create such a maze
        private void newMazeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            screenTimer.Stop();//stop any shenanigans from happening behind the dialog
            _winning = false;
            NewMazeForm newMaze = new NewMazeForm(_xSetSize, _ySetSize);//send current sizes to form
            newMaze.ShowDialog();
            //grab vars from form if result OK and build maze
            if (newMaze.DialogResult == DialogResult.OK)
            {
                _xSetSize = newMaze.VerticalSize;
                _ySetSize = newMaze.HorizontalSize;
                SummonMaze(_xSetSize, _ySetSize);
                _myMazeDrawing = new MazeDrawing(_mazeCells, _xSetSize, _ySetSize, _xStartPos, _yStartPos, _xFinishPos, _yFinishPos);
                gamePanel.Invalidate();
            }
        }

        //draw the maze
        private void gamePanel_Paint(object sender, PaintEventArgs e)
        {
            
            //find max scale for maze
            decimal xScaleMax = (gamePanel.Size.Height - 1) / _xSetSize;
            decimal yScaleMax = (gamePanel.Size.Width - 1) / _ySetSize;
            if (yScaleMax < xScaleMax)
            {
                xScaleMax = yScaleMax;
            }
            _myMazeDrawing.CellScale = Convert.ToInt32(xScaleMax);

            //left indent 1/2 dif between panel width and maze width
            int leftIndent = (int)((gamePanel.Size.Width - (xScaleMax * _ySetSize)) / 2);
            //top indent 1/2 dif between panel height and maze height
            int topIndent = (int)((gamePanel.Size.Height - (xScaleMax * _xSetSize)) / 2);

            if (leftIndent > 0)
            {
                _myMazeDrawing.LrIndent = leftIndent;
            }
            else
            {
                _myMazeDrawing.LrIndent = 0;
            }

            if (topIndent > 0)
            {
                _myMazeDrawing.UdIndent = topIndent;
            }
            else
            {
                _myMazeDrawing.UdIndent = 0;
            }

            //draw the maze!
            _myMazeDrawing.DrawAsGraphics(e, _winning);
        }

        // Export maze as text.
        // http://msdn.microsoft.com/en-us/library/sfezx97z.aspx
        private void textToolStripMenuItem_Click(object sender, EventArgs e)
        {
            screenTimer.Stop();//stop any shenanigans from happening behind the dialog
            saveFileDialog.FileName = string.Empty;
            saveFileDialog.Filter = "Unicode text file (*.txt)|*.txt";
            saveFileDialog.Title = "Save maze as text";

            saveFileDialog.ShowDialog();
            if (saveFileDialog.FileName != string.Empty)
            {
                try
                {
                    //Write the string to a file.
                    //http://msdn.microsoft.com/en-us/library/aa287548(v=vs.71).aspx
                    System.IO.StreamWriter outFile = new System.IO.StreamWriter(saveFileDialog.FileName);

                    List<string> mazeText = _myMazeDrawing.DrawAsText();//lines that represent the maze

                    foreach (string line in mazeText)
                    {
                        outFile.WriteLine(line);
                    }
                    outFile.Close();
                }
                catch (System.IO.IOException ioE)
                {
                    MessageBox.Show(ioE.Message);
                }
            }
        }

        // Export maze as image file (various raster types)
        private void imageToolStripMenuItem_Click(object sender, EventArgs e)
        {
            screenTimer.Stop();//stop any shenanigans from happening behind the dialog
            //Major TODO: SVG creation :-)
            saveFileDialog.FileName = string.Empty;
            saveFileDialog.Filter = "Portable Network Graphics (*.png)|*.png|Windows Bitmap (*.bmp)|*.bmp|Graphics Interchange Format (*.gif)|*.gif|JPEG Image (*.jpg)|*.jpg";
            saveFileDialog.Title = "Save maze as image";

            saveFileDialog.ShowDialog();
            if (saveFileDialog.FileName != string.Empty)
            {
                try
                {
                    
                    //Saves the Image via a FileStream created by the OpenFile method.
                    //http://msdn.microsoft.com/en-us/library/sfezx97z.aspx
                    System.IO.FileStream fs =
                       (System.IO.FileStream)saveFileDialog.OpenFile();

                    // Saves the Image in the appropriate ImageFormat based upon the
                    // File type selected in the dialog box.
                    // NOTE that the FilterIndex property is one-based.
                    switch (saveFileDialog.FilterIndex)
                    {
                        //case 1:
                          //{
                          //    MessageBox.Show("SVG coming soon."); //maybe
                          //    break;
                          //}

                        case 1:
                            {
                                System.Drawing.Bitmap mazeImage = _myMazeDrawing.DrawAsImage(25, 2);
                                mazeImage.Save(fs,
                                   System.Drawing.Imaging.ImageFormat.Png);
                                break;
                            }

                        case 2:
                            {
                                System.Drawing.Bitmap mazeImage = _myMazeDrawing.DrawAsImage(25, 2);
                                mazeImage.Save(fs,
                                   System.Drawing.Imaging.ImageFormat.Bmp);
                                break;
                            }

                        case 3:
                            {
                                System.Drawing.Bitmap mazeImage = _myMazeDrawing.DrawAsImage(25, 2);
                                mazeImage.Save(fs,
                                   System.Drawing.Imaging.ImageFormat.Gif);
                                break;
                            }

                        case 4:
                            {
                                System.Drawing.Bitmap mazeImage = _myMazeDrawing.DrawAsImage(25, 2);
                                mazeImage.Save(fs,
                                   System.Drawing.Imaging.ImageFormat.Jpeg);
                                break;
                            }
                    }

                    fs.Close();
                }
                catch (System.IO.IOException ioE)
                {
                    MessageBox.Show(ioE.Message);
                }
            }
        }

        //ensure maze gets resized when form does
        //http://blogs.msdn.com/b/alejacma/archive/2008/11/20/controls-won-t-get-resized-once-the-nesting-hierarchy-of-windows-exceeds-a-certain-depth-x64.aspx
        private void MaizeGameForm_Resize(object sender, EventArgs e)
        {
            if (gamePanel.Handle != null)
            {
                BeginInvoke(new MethodInvoker(ChangeSize));
            }
        }

        //called by form's resize event handler; resets & redraws the gamePanel
        protected void ChangeSize()
        {
            gamePanel.Invalidate();
            gamePanel.Update();
        }

        //restart the current maze
        private void restartMazeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            screenTimer.Stop();//stop any shenanigans from happening
            _winning = false;//clear "win" state
            //move the player back to the starting point:
            //take player away from where he is
            _mazeCells[_playerX, _playerY].PlayerMarker = false;
            //set his coords to the starting position
            _playerX = _xStartPos;
            _playerY = _yStartPos;
            //set that the player is there
            _mazeCells[_playerX, _playerY].PlayerMarker = true;

            //zap all the breadcrumbs
            for (int i = 0; i < _xSetSize; i++)
            {
                for (int j = 0; j < _ySetSize; j++)
                {
                    _mazeCells[i, j].Breadcrumb = false;
                }
            }

            //redraw the maze
            gamePanel.Invalidate();
            gamePanel.Update();
        }

        //listener for key presses - UDLR moves character, SPACE clears win text
        private void MaizeGameForm_KeyDown(object sender, KeyEventArgs e)
        {
            switch (e.KeyData.ToString())
            {
                case "Left":
                    if (!_winning && !_mazeCells[_playerX, _playerY].isWalled((int) Directions.W))
                    {
                        _mazeCells[_playerX, _playerY].PlayerMarker = false;
                        if (leaveBreadcrumbsMenuItem.Checked)
                        {
                            _mazeCells[_playerX, _playerY].Breadcrumb = true;                        
                        }
                        _playerY -= 1;
                        _mazeCells[_playerX, _playerY].PlayerMarker = true;
                    }
                    break;

                case "Right":
                    if (!_winning && !_mazeCells[_playerX, _playerY].isWalled((int)Directions.E))
                    {
                        _mazeCells[_playerX, _playerY].PlayerMarker = false;
                        if (leaveBreadcrumbsMenuItem.Checked)
                        {
                            _mazeCells[_playerX, _playerY].Breadcrumb = true;
                        }
                        _playerY += 1;
                        _mazeCells[_playerX, _playerY].PlayerMarker = true;
                    }
                    break;

                case "Up":
                    if (!_winning && !_mazeCells[_playerX, _playerY].isWalled((int)Directions.N))
                    {
                        _mazeCells[_playerX, _playerY].PlayerMarker = false;
                        if (leaveBreadcrumbsMenuItem.Checked)
                        {
                            _mazeCells[_playerX, _playerY].Breadcrumb = true;
                        }
                        _playerX -= 1;
                        _mazeCells[_playerX, _playerY].PlayerMarker = true;
                    }
                    break;

                case "Down":
                    if (!_winning && !_mazeCells[_playerX, _playerY].isWalled((int)Directions.S))
                    {
                        _mazeCells[_playerX, _playerY].PlayerMarker = false;
                        if (leaveBreadcrumbsMenuItem.Checked)
                        {
                            _mazeCells[_playerX, _playerY].Breadcrumb = true;
                        }
                        _playerX += 1;
                        _mazeCells[_playerX, _playerY].PlayerMarker = true;
                    }
                    break;

                case "Space": //clears win message, goes on to harder maze
                    if (_winning && screenTimer.Enabled)
                    {
                        screenTimer.Stop();
                        clearGameField();
                    }
                    break;

            }

            //when player moves into finish position, set "win" state
            if (!_winning && ((_playerX == _xFinishPos) && (_playerY == _yFinishPos)))
            {
                _winning = true;
                //show congratulatory message for 5 seconds
                screenTimer.Interval = 5000;
                screenTimer.Tick += new EventHandler(resetGameField);
                screenTimer.Start();
            }

            //reset, redraw
            gamePanel.Invalidate();
            gamePanel.Update();

        }

        //after player wins, summon a harder maze
        private void clearGameField()
        {
            //sanity check
            if (!_winning)
            {
                return;            
            }
            //clear win state
            _winning = false;

            //call up "50%" harder maze
            _xSetSize += (_xSetSize / 2);
            _ySetSize += (_ySetSize / 2);
            if (_xSetSize > 300)
            {
                _xSetSize = 300;
            }
            if (_ySetSize > 300)
            {
                _ySetSize = 300;
            }
            SummonMaze(_xSetSize, _ySetSize);

            //reset, redraw
            gamePanel.Invalidate();
            gamePanel.Update();
        }

        //called by screen timer after time elapsed after win
        private void resetGameField(object sender, EventArgs e)
        {
            screenTimer.Stop();
            clearGameField();
        }
    }
}
